"""API methods for working with single resources."""
import json

import requests

from cmem.cmempy import config
from cmem.cmempy.api import (
    request,
    send_request
)


def get_resource_uri():
    """Get endpoint URI pattern for a resource."""
    path = "/workspace/projects/{}/resources/{}"
    return config.get_di_api_endpoint() + path


def get_metadata_uri():
    """Get metadata endpoint URI pattern for a resource."""
    path = "/workspace/projects/{}/resources/{}/metadata"
    return config.get_di_api_endpoint() + path


def get_resource(project_name, resource_name):
    """GET resource.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        requests.Response object
    """
    response = send_request(
        get_resource_uri().format(project_name, resource_name),
        method="GET"
    )
    return response


def resource_exist(project_name, resource_name):
    """Check if a resource exist.

    A return value of true means the resource exists. A return value of true
    means the resource does not exists.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        bool
    """
    try:
        get_resource_metadata(project_name, resource_name)
        return True
    except requests.exceptions.HTTPError as error:
        if error.response.status_code == 404:
            return False
        raise error


def get_resource_response(project_name, resource_name):
    """GET resource as streamable request response.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        requests.Response object
    """
    resource_url = get_resource_uri().format(project_name, resource_name)
    return request(
        resource_url,
        method="GET",
        stream=True
    )


def make_new_resource(project_name, resource_name, data=None, files=None):
    """PUT create new resource.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.
        data (dict): Dictionary (see requests.request) for the body
        files (dict): Dictionary (see requests.request) for multipart upload
    """
    send_request(
        get_resource_uri().format(project_name, resource_name),
        method="PUT",
        data=data,
        files=files
    )


def create_resource(
        project_name,
        resource_name,
        file_resource=None,
        replace=False
):
    """Create a new resource (streamed).

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.
        file_resource (file stream): Already opened byte file stream
        replace (bool): Replace resource if needed.

    Returns:
        requests.Response object

    Raises:
        ValueError: missing parameter
        ValueError: Resource exist and no replace enabled
    """
    if not file_resource:
        raise ValueError("Parameter file_name is needed.")
    if not replace and resource_exist(project_name, resource_name):
        raise ValueError(
            "Resource {} already exists in project {}."
            .format(resource_name, project_name)
        )
    # https://requests.readthedocs.io/en/latest/user/advanced/#streaming-uploads
    with file_resource as file:
        response = request(
            get_resource_uri().format(project_name, resource_name),
            method="PUT",
            stream=True,
            data=file,
        )
    return response


def delete_resource(project_name, resource_name):
    """DELETE remove existing resource.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.
    """
    send_request(
        get_resource_uri().format(project_name, resource_name),
        method="DELETE"
    )


def get_resource_metadata(project_name, resource_name):
    """GET retrieve resource metadata.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        Depends on what json.loads gives back
    """
    response = send_request(
        get_metadata_uri().format(project_name, resource_name),
        method="GET"
    )
    return json.loads(response.decode("utf-8"))


def get_resource_usage_data(project_name, resource_name):
    """GET retrieve the usage data of a resource.

    Args:
        project_name (str): The project ID in the workspace.
        resource_name (str): The resource ID/name in the workspace.

    Returns:
        Depends on what json.loads gives back
    """
    response = send_request(
        get_resource_uri().format(project_name, resource_name) + "/usage",
        method="GET"
    )
    return json.loads(response.decode("utf-8"))



"""API for eccenca DataPlatform /api/admin/backup/*."""

from requests_toolbelt.multipart import encoder

from cmem.cmempy import config
from cmem.cmempy.api import request


def get_zip():
    """Get a full backup of the knowledge graphs."""
    uri = config.get_dp_api_endpoint() + "/api/admin/backup/zip"
    return request(
        uri,
        method="GET",
        stream=True
    )


def post_zip(zip_file):
    """Post a full backup of the knowledge graphs."""
    uri = config.get_dp_api_endpoint() + "/api/admin/restore/zip"
    with open(zip_file, 'rb') as file:
        files = [
            (
                'file',
                ('backup.zip', file, 'application/zip')
            )
        ]
        # https://stackoverflow.com/questions/42281387/
        multipart_encoder = encoder.MultipartEncoder(
            fields=files,
            boundary="xxx---------------xxx",
        )
        return request(
            uri,
            method="POST",
            data=multipart_encoder,
            headers={'Content-Type': multipart_encoder.content_type},
            stream=True
        )



"""API for working with single transformation rule."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_transformation_rule_uri():
    """Get endpoint URI pattern for a transformation task rule."""
    path = "/transform/tasks/{}/{}/rule/{}"
    return config.get_di_api_endpoint() + path


def get_transformation_rule(project_name, task_name, rule_name):
    """GET transformation rule for the task by name."""
    headers = {"Accept": "application/json"}
    response = send_request(
        get_transformation_rule_uri().format(project_name,
                                             task_name,
                                             rule_name),
        method="GET",
        headers=headers
    )
    return json.loads(response.decode("utf-8"))


def update_transformation_rule(project_name, task_name, rule_name, data=None):
    """PUT update transformation rule for the task."""
    headers = {"Content-Type": "application/json"}
    send_request(
        get_transformation_rule_uri().format(project_name,
                                             task_name,
                                             rule_name),
        method="PUT",
        data=data,
        headers=headers
    )



"""API methods for working with one project."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_projects_uri():
    """Get endpoint URI for project list."""
    path = "/workspace/projects"
    return config.get_di_api_endpoint() + path


def get_project_uri():
    """Get endpoint URI pattern for a project."""
    path = "/workspace/projects/{}"
    return config.get_di_api_endpoint() + path


def get_project_api_uri():
    """Get endpoint URI pattern for a project (new API)."""
    path = "/api/workspace/projects/{}"
    return config.get_di_api_endpoint() + path


def get_projects():
    """GET all projects."""
    response = send_request(get_projects_uri(),
                            method="GET")
    return json.loads(response.decode("utf-8"))


def get_project(name):
    """GET one project."""
    response = send_request(get_project_uri().format(name),
                            method="GET")
    return json.loads(response.decode("utf-8"))


def reload_project(name):
    """Reload one project."""
    return send_request(
        get_project_uri().format(name) + "/reload",
        method="POST"
    )


def get_prefixes(name):
    """GET prefixes of a project."""
    response = send_request(
        get_project_api_uri().format(name) + "/prefixes",
        method="GET",
        headers={"Accept": "application/json"}
    )
    return json.loads(response.decode("utf-8"))


def make_new_project(name):
    """PUT make new project."""
    response = send_request(get_project_uri().format(name),
                            method="PUT")
    return json.loads(response.decode("utf-8"))


def delete_project(name):
    """DELETE remove project."""
    send_request(get_project_uri().format(name),
                 method="DELETE")


def get_failed_tasks_report(name):
    """Get a report of tasks that could not be loaded in a project."""
    response = send_request(
        get_project_api_uri().format(name) + "/failedTasksReport",
        method="GET",
        headers={"Accept": "application/json"}
    )
    return json.loads(response.decode("utf-8"))



"""API for searching items in the data integration workspace."""

import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_search_items_uri():
    """Get endpoint URL for the search items API."""
    path = "/api/workspace/searchItems"
    return config.get_di_api_endpoint() + path


# pylint: disable-msg=too-many-arguments
def list_items(
        limit=1000000,
        offset=0,
        project=None,
        text_query="",
        item_type="",
        facets=None,
        add_task_parameter=True
):
    """GET retrieve custom task for the project."""
    data = json.dumps({
        "limit": limit,
        "offset": offset,
        "itemType": item_type,
        "project": project,
        "textQuery": text_query,
        "facets": facets,
        "addTaskParameters": add_task_parameter
    })
    response = send_request(
        get_search_items_uri(),
        method="POST",
        headers={'Content-Type': 'application/json'},
        data=data
    )
    return json.loads(response.decode("utf-8"))



"""API for working with activities."""

ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW = "ExecuteDefaultWorkflow"

ACTIVITY_TYPE_EXECUTE_LOCALWORKFLOW = "ExecuteLocalWorkflow"

VALID_ACTIVITY_STATUS = [
    "Idle",
    "Not executed",
    "Finished",
    "Cancelled",
    "Failed",
    "Successful",
    "Canceling",
    "Running",
    "Waiting"
]



"""API for getting tasks in the data integration workspace."""

import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_task_uri():
    """Get endpoint URL for the tasks API."""
    path = "/workspace/projects/{}/tasks/{}"
    return config.get_di_api_endpoint() + path


def get_task(
        project=None,
        task=None,
        with_labels=True
):
    """GET a task description."""
    response = send_request(
        get_task_uri().format(
            project,
            task
        ) + "?withLabels=" + str(with_labels).lower(),
        method="GET",
        headers={'Accept': 'application/json'},
    )
    return json.loads(response.decode("utf-8"))


def delete_task(
        project=None,
        task=None
):
    """DELETE a task."""
    send_request(
        get_task_uri().format(
            project,
            task
        ),
        method="DELETE"
    )


def patch_parameter(
        project=None,
        task=None,
        data=None
):
    """PATCH a task."""
    headers = {'Content-Type': 'application/json'}
    data = json.dumps(data)
    send_request(
        get_task_uri().format(
            project,
            task
        ),
        data=data,
        headers=headers,
        method="PATCH"
    )



"""API for working with several workflows."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_workflow_uri():
    """Get endpoint URI pattern for a workflow."""
    path = "/workflow/workflows/{}"
    return config.get_di_api_endpoint() + path


def get_workflows(project_name):
    """GET retrieve list of workflows."""
    response = send_request(get_workflow_uri().format(project_name),
                            method="GET")
    return json.loads(response.decode("utf-8"))



"""API for health and version information."""
import json
import re

from requests import HTTPError

from cmem.cmempy import config
from cmem.cmempy.api import get_json, send_request
from cmem.cmempy.queries import SparqlQuery

SHAPE_CATALOG_VERSION_QUERY = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX : <https://vocab.eccenca.com/shacl/>
SELECT ?version
FROM :
WHERE {
  : owl:versionInfo ?version
}
ORDER BY ASC(?version)
"""


def get_dp_health_endpoint():
    """Get DataPlatform health endpoint."""
    return config.get_dp_api_endpoint() + "/actuator/health"


def get_dp_info_endpoint():
    """Get DataPlatform version endpoint."""
    return config.get_dp_api_endpoint() + "/actuator/info"


def get_di_health_endpoint():
    """Get DataPlatform health endpoint."""
    return config.get_di_api_endpoint() + "/health"


def get_di_version_endpoint():
    """Get DataPlatform version endpoint."""
    return config.get_di_api_endpoint() + "/version"


def get_dm_version_endpoint():
    """Get DataManager version endpoint."""
    return config.get_cmem_base_uri() + "/version.html"


def get_shape_catalog_version():
    """GET version of the ShapeCatalog."""
    try:
        results = SparqlQuery(
            SHAPE_CATALOG_VERSION_QUERY,
        ).get_json_results()["results"]["bindings"]
        if len(results) > 0:
            return str(results[0]["version"]["value"])
    except (KeyError, HTTPError):
        pass
    return "UNKNOWN"


def get_dm_version():
    """GET version of DataManager."""
    url = get_dm_version_endpoint()
    try:
        response = send_request(url).decode()
    except HTTPError:
        return "ERROR"
    try:
        response = re.findall(r'v[0-9]+\..*', response)[0]
        return response
    except Exception:  # pylint: disable=broad-except
        return "UNKNOWN"


def get_dp_version():
    """GET version of DataPlatform."""
    response = None
    url = get_dp_info_endpoint()
    try:
        response = send_request(url)
    except Exception:  # pylint: disable=broad-except
        response = None
        # TODO: checking health status needs to be improved
    if response is None:
        url = url.replace("/actuator", "")
        response = send_request(url)
    result = json.loads(response)
    return result["version"]


def dp_is_healthy():
    """Check status of DataIntegration."""
    url = get_dp_health_endpoint()
    try:
        response = send_request(url)
    except Exception:  # pylint: disable=broad-except
        response = None
        # TODO: checking health status needs to be improved
    if response is None:
        url = url.replace("/actuator", "")
        response = send_request(url)
    result = json.loads(response)
    if result['status'] == "UP":
        return True
    return False


def get_di_version():
    """GET version of DataIntegration."""
    response = send_request(get_di_version_endpoint())
    return response.decode("utf-8")


def di_is_healthy():
    """Check status of DataIntegration."""
    try:
        result = json.loads(send_request(get_di_health_endpoint()))
    except ValueError:
        return False
    if result['status'] == "UP":
        return True
    return False


def get_complete_status_info() -> dict:  # noqa: C901
    """Collect all status info and provide proper dictionary."""
    info = {
        "di": {
            "version": "UNKNOWN",
            "healthy": "UNKNOWN"
        },
        "dp": {
            "version": "UNKNOWN",
            "healthy": "UNKNOWN"
        },
        "dm": {
            "version": "UNKNOWN",
            "healthy": "UNKNOWN"
        },
        "shapes": {
            "version": "UNKNOWN",
            "healthy": "UNKNOWN"
        },
        "store": {
            "type": "STORE",
            "version": "UNKNOWN",
            "healthy": "UNKNOWN"
        },
        "overall": {
            "healthy": "UNKNOWN"
        }
    }
    e_prefix = "Error: "
    try:
        info["di"]["version"] = get_di_version()
        info["di"]["healthy"] = "UP" if di_is_healthy() else "DOWN"
    except Exception as error:  # pylint: disable=broad-except
        info["di"]["error"] = e_prefix + str(error)
        info["di"]["healthy"] = "DOWN"
    try:
        info["dp"]["info"] = get_json(get_dp_info_endpoint())
        info["dp"]["health"] = get_json(get_dp_health_endpoint())
        info["dp"]["version"] = info["dp"]["info"]["version"]
        info["dp"]["healthy"] = info["dp"]["health"]["status"]
        if "store" in info["dp"]["info"]:
            info["store"]["type"] = info["dp"]["info"]["store"]["type"]
            info["store"]["version"] = info["dp"]["info"]["store"]["version"]
        info["store"]["healthy"] = info["dp"]["healthy"]
    except Exception as error:  # pylint: disable=broad-except
        info["dp"]["error"] = e_prefix + str(error)
        info["dp"]["healthy"] = "DOWN"
    try:
        dm_version = get_dm_version()
        info["dm"]["version"] = \
            dm_version if dm_version != "ERROR" else "UNKNOWN"
        info["dm"]["healthy"] = "UP" if dm_version != "ERROR" else "DOWN"
    except Exception as error:  # pylint: disable=broad-except
        info["dm"]["error"] = e_prefix + str(error)
        info["dm"]["healthy"] = "DOWN"
    try:
        info["shapes"]["version"] = get_shape_catalog_version()
        if info["shapes"]["version"] == "UNKNOWN":
            info["shapes"]["healthy"] = "DOWN"
        else:
            info["shapes"]["healthy"] = "UP"
    except Exception as error:  # pylint: disable=broad-except
        info["shapes"]["error"] = e_prefix + str(error)
    # combine overall.status
    healthy_string =\
        info["dp"]["healthy"] + info["store"]["healthy"] + \
        info["di"]["healthy"] + info["shapes"]["healthy"]
    if healthy_string == "UPUPUPUP":  # 🙈
        info["overall"]["healthy"] = "UP"
    if "DOWN" in healthy_string:
        info["overall"]["healthy"] = "DOWN"
    return info



"""API for DP /proxy/{} endpoint."""

import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get(endpoint_id="default"):
    """Retrieve list of graphs from DP endpoint."""
    proxy_uri = config.get_dp_api_endpoint() + "/proxy/{}"
    return json.loads(send_request(
        proxy_uri.format(
            endpoint_id
        ),
        method="GET"
    ).decode("utf-8"))



"""API for eccenca DataPlatform Label Resolution."""

import json

from cmem.cmempy import config
from cmem.cmempy.api import request


def resolve(iris, graph=None):
    """Resolve IRIs to short form labels.

    If none can be found in the data, the iri itself is used to create one.

    Possible extension: langPrefs parameter

    Args:
        iris: The list of iris to resolve.
        graph: The context/graph, in which the lookup should take place.

    Raises:
        ValueError
    """
    uri = config.get_dp_api_endpoint() + "/api/explore/titles"

    if not isinstance(iris, list):
        raise ValueError(
            "Parameter 'iris' has to be a list of (iri) strings."
        )

    data = json.dumps(iris)
    headers = {
        "Content-type": "application/json",
        "Accept": "application/json"
    }

    params = {}
    if graph is not None:
        params["contextGraph"] = graph

    response = json.loads(request(
        uri,
        method="POST",
        headers=headers,
        params=params,
        data=data
    ).content)
    return response



"""API methods for working with a single workflow."""
from cmem.cmempy import config
from cmem.cmempy.api import (
    get_json,
    request,
    send_request
)


def get_workflow_uri():
    """Get endpoint URI pattern for a workflow."""
    return config.get_di_api_endpoint() + "/workflow/workflows/{}/{}"


def get_workflow_execute_uri():
    """Get execute URI pattern for a workflow."""
    return get_workflow_uri() + "/execute"


def get_workflow_status_uri():
    """Get status URI pattern for a workflow."""
    return get_workflow_uri() + "/status"


def get_workflow_execute_on_payload_uri():
    """Get execute on payload URI pattern for a workflow."""
    return get_workflow_uri() + "/executeOnPayload"


def get_workflow_editor_uri():
    """Get execute URI pattern for a workflow.

    Since 21.01 this is changed to the new workbench
    """
    return config.get_di_api_endpoint() + "/workbench/projects/{}/workflow/{}"


def get_workflow(project_name, task_name):
    """GET retrieve single workflow."""
    return send_request(
        get_workflow_uri().format(project_name, task_name),
        method="GET"
    ).decode("utf-8")


def make_new_workflow(project_name, task_name, data=None):
    """PUT create workflow."""
    headers = {"Content-Type": "application/xml"}
    return send_request(
        get_workflow_uri().format(project_name, task_name),
        method="PUT",
        data=data,
        headers=headers
    ).decode("utf-8")


def execute_on_payload(project_name, task_name, data=None,
                       headers=None):
    """POST execute on payload."""
    if not headers:
        headers = {"Content-Type": "application/xml"}

    # pylint: disable-msg=duplicate-code
    return send_request(
        get_workflow_execute_on_payload_uri().format(project_name, task_name),
        method="POST",
        data=data,
        headers=headers
    ).decode("utf-8")


def delete_workflow(project_name, task_name):
    """DELETE remove workflow."""
    return send_request(
        get_workflow_uri().format(project_name, task_name),
        method="DELETE"
    ).decode("utf-8")


def execute_workflow(project_name, task_name):
    """PUT execute workflow."""
    return send_request(
        get_workflow_execute_uri().format(project_name, task_name),
        method="PUT"
    ).decode("utf-8")


def get_workflow_status(project_name, task_name):
    """GET workflow status.

    Returns gibberish, internal API endpoint
    use cmempy.workspace.activities.taskactivities.get_activity_status
    instead
    """
    return send_request(
        get_workflow_status_uri().format(project_name, task_name),
        method="GET"
    ).decode("utf-8")


def get_workflows_io():
    """Get a list of suitable io workflow incl. input/output information.

    see CMEM-3089

    Args:
        project_name: project ID

    Returns:
        list of dicts of workflow infos
    """
    endpoint = "{}/api/workflow/info".format(config.get_di_api_endpoint())
    params = {}
    io_workflows = []
    for _ in get_json(endpoint, method="GET", params=params):
        ins = len(_["variableInputs"])
        outs = len(_["variableOutputs"])
        if ins == 1 or outs == 1:
            io_workflows.append(_)
    return io_workflows


# pylint: disable=too-many-arguments
def execute_workflow_io(
        project_name,
        task_name,
        input_file=None,
        input_mime_type="application/xml",
        output_mime_type="application/xml",
        auto_config=False
):
    """Execute a workflow with variable input or output from or to a file.

    see CMEM-3089

    Args:
        project_name: project ID
        task_name: workflow ID
        input_file: file path
        input_mime_type: A mime type string
        output_mime_type: A mime type string
        auto_config (bool): enable autoconfiguration of input dataset

    Returns:
        Response object from requests API
    """
    endpoint = "{}/api/workflow/result/{}/{}".format(
        config.get_di_api_endpoint(), project_name, task_name
    )
    if auto_config:
        endpoint = f"{endpoint}?config-general-autoConfig=true"
    headers = {
        "Accept": output_mime_type
    }
    if input_file:
        # in case an input is given do a streaming upload
        headers["Content-Type"] = input_mime_type
        with open(input_file, 'rb') as data:
            return request(
                endpoint,
                method="POST",
                headers=headers,
                data=data,
                stream=True
            )
    # without an input, just a post without a body
    return request(
        endpoint,
        method="POST",
        headers=headers,
        stream=True
    )



"""API for running SPARQL queries on DP proxy endpoints."""

from base64 import b64encode

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_update_endpoint():
    """Get endpoint URI pattern for a SPARQL update endpoint."""
    return config.get_dp_api_endpoint() + "/proxy/{}/update"


# pylint: disable-msg=too-many-arguments
def post(query,
         endpoint_id="default",
         accept="application/sparql-update",
         named_graph_uri=(),
         default_graph_uri=(),
         base64_encoded=False
         ):
    """POST update / insert query."""
    headers = {"Accept": accept}
    uri = get_update_endpoint().format(endpoint_id)

    if base64_encoded:
        query = b64encode(query.encode())
        data = {'update': query, 'base64encoded': 'true'}
    else:
        data = {'update': query}

    if named_graph_uri:
        data['named-graph-uri'] = []
        for graph in named_graph_uri:
            data['named-graph-uri'].append(graph)

    if default_graph_uri:
        data['default-graph-uri'] = []
        for graph in default_graph_uri:
            data['default-graph-uri'].append(graph)

    response = send_request(
        uri=uri,
        method="POST",
        data=data,
        headers=headers
    )
    return response.decode("utf-8")



"""API for running SPARQL queries on DP proxy endpoints."""

from base64 import b64encode

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_query_api_endpoint():
    """Get endpoint URI pattern for a SPARQL query endpoint."""
    return config.get_dp_api_endpoint() + "/proxy/{}/sparql"


# pylint: disable-msg=too-many-arguments
def get(
        query,
        endpoint_id="default",
        owl_imports_resolution=True,
        accept="application/sparql-results+json",
        named_graph_uri=(),
        default_graph_uri=(),
        base64_encoded=False,
        cd_target=None
):
    """GET query results.

    Args:
        query (str): The SPARQL query.
        endpoint_id: The DataPlatform endpoint ID
        owl_imports_resolution (bool): turn on owl:imports resolution,
            which means if a graph imports a second graph, the data of
            the second graph is included when querying the first graph
        accept (str): mime type, which content is accepted
        named_graph_uri (list): list of GRAPH URIs
        default_graph_uri (list):  list of GRAPH URIs
        base64_encoded (bool): flag to turn on base64 encoding of the request
        cd_target (str or None): Which name has the downloaded file?
    """
    headers = {"Accept": accept}
    uri = get_query_api_endpoint().format(endpoint_id)
    params = {}
    if base64_encoded:
        params["query"] = b64encode(query.encode())
        params["base64encoded"] = "true"
    else:
        params["query"] = query
    if cd_target:
        params["cdTarget"] = cd_target
    if len(named_graph_uri) > 0:
        params["named-graph-uri"] = named_graph_uri
    if len(default_graph_uri) > 0:
        params["default-graph-uri"] = default_graph_uri
    if not owl_imports_resolution:
        # default of owl_imports_resolution is true, so only add on false
        params["owlImportsResolution"] = "false"
    return send_request(
        uri, method="GET", headers=headers, params=params
    ).decode("utf-8")


def post(
        query,
        endpoint_id="default",
        owl_imports_resolution=True,
        accept="application/sparql-results+json",
        named_graph_uri=(),
        default_graph_uri=(),
        base64_encoded=False,
        cd_target=None,
        distinct=False,
        limit=None,
        offset=None,
        timeout=None
):
    """Get query results via POST request (recommended).

    Args:
        query (str): The SPARQL query.
        endpoint_id: The DataPlatform endpoint ID
        owl_imports_resolution (bool): turn on owl:imports resolution,
            which means if a graph imports a second graph, the data of
            the second graph is included when querying the first graph
        accept (str): mime type, which content is accepted
        named_graph_uri (list): list of GRAPH URIs
        default_graph_uri (list):  list of GRAPH URIs
        base64_encoded (bool): flag to turn on base64 encoding of the request
        cd_target (str or None): Which name has the downloaded file?
        distinct (bool): query override - turn on DISTINCT SELECT
        limit (int or None): query override - set SELECT LIMIT
        offset (int or None): query override - set SELECT OFFSET
        timeout (int or None): timeout in milliseconds
    """
    headers = {"Accept": accept}
    uri = get_query_api_endpoint().format(endpoint_id)
    data = {}
    if base64_encoded:
        data["query"] = b64encode(query.encode())
        data["base64encoded"] = "true"
    else:
        data["query"] = query
    if cd_target:
        data["cdTarget"] = cd_target
    if len(named_graph_uri) > 0:
        data["named-graph-uri"] = named_graph_uri
    if len(default_graph_uri) > 0:
        data["default-graph-uri"] = default_graph_uri
    if not owl_imports_resolution:
        # default of owl_imports_resolution is true, so only add on false
        data["owlImportsResolution"] = "false"
    if distinct:
        data["distinct"] = "true"
    if limit:
        data["limit"] = limit
    if offset:
        data["offset"] = offset
    if timeout:
        data["timeout"] = timeout
    return send_request(
        uri, method="POST", headers=headers, data=data
    ).decode("utf-8")



"""API for working with the query catalog."""
import json
import os
import re
from uuid import uuid1

from pyparsing import ParseException

from rdflib.plugins.sparql.parser import (
    parseUpdate
)
from rdflib.plugins.sparql import (
    prepareQuery
)

from cmem.cmempy.api import get_json, request
from cmem.cmempy.config import (
    get_cmem_base_uri,
    get_dp_api_endpoint
)
from cmem.cmempy.dp.proxy import sparql
from cmem.cmempy.dp.proxy import update

# read queries are send to the normal query endpoint
QUERY_TYPES_READ = ("SELECT", "ASK", "DESCRIBE", "CONSTRUCT")
# update queries are send to the update query endpoint
QUERY_TYPES_UPDATE_TRIPLE = ("UPDATE", "DELETE", "INSERT", "CLEAR", "LOAD")
QUERY_TYPES_UPDATE_GRAPHS = ("CREATE", "DROP", "COPY", "MOVE", "ADD")
QUERY_TYPES_UPDATE = QUERY_TYPES_UPDATE_TRIPLE + QUERY_TYPES_UPDATE_GRAPHS
# broken types
QUERY_TYPES_FAULTY = ("FAULTY", "UNKNOWN")
# all types
QUERY_TYPES = QUERY_TYPES_READ + QUERY_TYPES_UPDATE + QUERY_TYPES_FAULTY

QUERY_STRING = """
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX dct: <http://purl.org/dc/terms/>
    PREFIX shui: <https://vocab.eccenca.com/shui/>
    PREFIX dct: <http://purl.org/dc/terms/>
    SELECT DISTINCT ?query ?label ?text ?type ?description
    FROM <https://ns.eccenca.com/data/queries/>
    WHERE {
        ?query a shui:SparqlQuery ;
            rdfs:label ?label ;
            shui:queryText ?text ;
            shui:queryType ?type .
            OPTIONAL {?query dct:description ?description .}
    }
    """

DEFAULT_NS = "https://ns.eccenca.com/data/queries/"


def get_query_status():
    """Get status information of run and running queries."""
    endpoint = get_dp_api_endpoint() + "/api/admin/currentQueries"
    return get_json(endpoint)


def cancel_query(query_id: str):
    """Cancel a running query."""
    endpoint = f"{get_dp_api_endpoint()}/api/admin/currentQueries/{query_id}"
    return request(endpoint, method="DELETE")


def get_query_editor_for_uri():
    """Get query editor URI pattern for a query URI."""
    return get_cmem_base_uri() + "/query?tab=query&query={}"


def get_query_editor_for_string():
    """Get query editor URI pattern for a query string."""
    return get_cmem_base_uri() + "/query?tab=query&queryString={}"


class SparqlQuery:
    """A SPARQL query with optional placeholders."""

    # pylint: disable=too-many-arguments, too-many-instance-attributes

    def __init__(
            self, text, url=None, label=None, query_type=None,
            description=None, origin="unknown", placeholder=None):
        """Initialize a SparqlQuery object."""
        self.text = text
        self.short_url = None  # will be set with self.set_url
        self.url = self.set_url(url)
        self.label = label
        self.query_type = self.set_query_type(
            query_type, placeholder=placeholder
        )
        self.description = description
        # can be one of unknown, remote or file (used for get_editor_uri)
        self.origin = origin
        self.placeholder = placeholder

    def set_url(self, url=None):
        """Set (or generate) an URL for the query."""
        if url is not None:
            self.url = url
            self.short_url = url.replace(DEFAULT_NS, ':')
        else:
            uuid = str(uuid1())
            self.url = DEFAULT_NS + uuid
            self.short_url = ":" + uuid
        return self.url

    def __str__(self):
        """Get string representation (URL)."""
        return self.url

    def set_query_type(self, query_type=None, placeholder=None):
        """Set query type."""
        self.query_type = "UNKNOWN"
        if str(query_type).upper() in QUERY_TYPES:
            self.query_type = str(query_type).upper()
        else:
            parsed_type = str(
                self.get_query_type(placeholder=placeholder)
            ).upper()
            if parsed_type in QUERY_TYPES:
                self.query_type = parsed_type
        return self.query_type

    def get_query_type(self, placeholder=None):
        """Get type of query when placeholder are filled.

        Tries to get the type of the query - returns UNKNOWN if not possible,
        UPDATE, when an update query or a more specific READ query type
        """
        if placeholder is None:
            placeholder = {}
        algebra_names = {
            "DescribeQuery": "DESCRIBE",
            "ConstructQuery": "CONSTRUCT",
            "SelectQuery": "SELECT",
            "AskQuery": "ASK"
        }
        # prepare query text to parse
        string = self.text
        try:
            # try to fill with placeholders
            string = self.get_filled_text(placeholder)
        except ValueError:
            pass
        # 1. try parse as SPARQL query first and map algebra type to our type
        try:
            parsed_query = prepareQuery(string)
            if str(parsed_query.algebra.name) in algebra_names:
                # return the mapped read query type
                return algebra_names[str(parsed_query.algebra.name)]
        except ParseException:
            pass
        # 2. try to parse as UPDATE query
        try:
            parseUpdate(string)
            # return the default UPDATE query type
            return "UPDATE"
        except ParseException:
            pass
        # return the UNKNOWN query type
        return "UNKNOWN"

    def get_default_accept_header(self):
        """Return the default accept header string for the query.

        return value is based on the query type and biased towards
        command line shell environment.
        """
        default_header = {
            "SELECT": "text/csv",
            "ASK": "text/csv",
            "DESCRIBE": "text/turtle",
            "CONSTRUCT": "text/turtle"
        }
        return default_header.get(self.query_type, "*")

    def get_placeholder_keys(self, text=None):
        """Get all placeholder of a query text as a set of keys."""
        if text is None:
            text = self.text
        keys = re.findall(r"{{([a-zA-Z0-9_-]+)}}", text)
        return set(keys)

    def get_filled_text(self, placeholder):
        """Replace placeholders with given values and return text.

        raises an ValueError exception if not all placeholders are filled.
        """
        text = self.text
        for key, value in placeholder.items():
            text = text.replace('{{' + key + '}}', value)
        if self.get_placeholder_keys(text):
            raise ValueError("Not all placeholders filled for executing the "
                             "following query:\n"
                             "- Label: " + str(self.label) + "\n"
                             "- ID: " + str(self.url) + "\n"
                             "- Missing parameters: {}"
                             .format(self.get_placeholder_keys(text)))
        return text

    def get_csv_results(self, placeholder=None, owl_imports_resolution=True):
        """Get results as CSV text."""
        if self.query_type != "SELECT":
            raise ValueError("Wrong query type. "
                             "CSV result only supported for SELECT queries.")
        results = self.get_results(
            placeholder=placeholder,
            owl_imports_resolution=owl_imports_resolution,
            accept="text/csv"
        )
        return results

    def get_json_results(self, placeholder=None, owl_imports_resolution=True):
        """Get results as parsed json object."""
        results = self.get_results(
            placeholder=placeholder,
            owl_imports_resolution=owl_imports_resolution,
            accept="application/sparql-results+json"
        )
        return json.loads(results)

    def get_results(self, placeholder=None,
                    owl_imports_resolution=True,
                    base64_encoded=False,
                    accept="application/sparql-results+json",
                    distinct=False,
                    limit=None,
                    offset=None,
                    timeout=None,
                    replace_placeholder=True
                    ):
        """Get results as raw output from cmem.cmempy.dp.proxy.sparql."""
        if placeholder is None:
            placeholder = {}
        if self.query_type is None:
            # try to do last minute determination of type
            self.query_type = self.get_query_type(placeholder)
        if replace_placeholder:
            query_text = self.get_filled_text(placeholder)
        else:
            query_text = self.text
        if self.query_type in QUERY_TYPES_UPDATE:
            return update.post(
                query=query_text,
                accept=accept,
                base64_encoded=base64_encoded
            )
        return sparql.post(
            query=query_text,
            accept=accept,
            base64_encoded=base64_encoded,
            owl_imports_resolution=owl_imports_resolution,
            distinct=distinct,
            limit=limit,
            offset=offset,
            timeout=timeout
        )

    def get_editor_url(self):
        """Get query editor URI for the query."""
        if self.origin == "remote":
            return get_query_editor_for_uri().format(self.url)
        if self.origin == "file":
            return get_query_editor_for_string().format(self.text)
        raise ValueError(
            "Unknown origin, can not provide editor URL: " + self.origin
        )


class QueryCatalog:
    """A representation of the query catalog."""

    def __init__(self):
        """Initialize the catalog."""
        self.queries = None

    def get_query(self, identifier, placeholder=None):
        """Get a query by giving an url, a short_url or a file name."""
        if os.path.isfile(identifier):
            # do not fetch the catalog in case a file is requested
            with open(identifier, "rU", encoding="UTF-8") as file_handle:
                return SparqlQuery(
                    file_handle.read(),
                    label="query from file " + identifier,
                    origin="file",
                    placeholder=placeholder
                )
        queries = self.get_queries()
        if DEFAULT_NS + identifier[1:] in queries:
            # if identifier comes from shell (:uuid) then we remove : and add
            # the default namespace
            return queries[DEFAULT_NS + identifier[1:]]
        if identifier in queries:
            # maybe the correct full URI is given
            return queries[identifier]
        return None

    def get_queries(self):
        """Get the query catalog as a dictionary of SparqlQuery objects."""
        if self.queries is not None:
            return self.queries
        queries = {}
        results = SparqlQuery(
            QUERY_STRING,
            query_type="SELECT"
        ).get_json_results()
        for query in results["results"]["bindings"]:
            description = None
            if description in query:
                description = query["description"]
            queries[query["query"]["value"]] = SparqlQuery(
                query["text"]["value"],
                url=query["query"]["value"],
                label=query["label"]["value"],
                query_type=query["type"]["value"],
                description=description,
                origin="remote"
            )
        self.queries = queries
        return self.queries


QUERY_CATALOG = QueryCatalog()



"""Import workspace."""
import os

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_import_uri():
    """Get endpoint URI pattern for workspace import."""
    return config.get_di_api_endpoint() + "/workspace/import/{}"


def import_xml_zip(xml_zip_file_path):
    """Import DI Workspace from XML ZIP."""
    with open(xml_zip_file_path, 'rb') as file_handle:
        xml_zip_file = {
            'file': (
                os.path.basename(xml_zip_file_path),
                file_handle
            )
        }
        import_uri = get_import_uri().format("xmlZip")
        response = send_request(
            import_uri,
            method="POST",
            files=xml_zip_file
        )
        return response


# https://jira.brox.de/browse/CMEM-2144
def import_rdf_turtle(rdf_turtle_zip_file_path):
    """Import DI Workspace from RDF Turtle (without resource files)."""
    with open(rdf_turtle_zip_file_path, 'rb') as file_handle:
        rdf_turtle_zip_file = {
            'file': (
                os.path.basename(rdf_turtle_zip_file_path),
                file_handle
            )
        }
        import_uri = get_import_uri().format("rdfTurtle")
        response = send_request(
            import_uri,
            method="POST",
            files=rdf_turtle_zip_file
        )
        return response


def import_workspace(file_path, plugin_name="xmlZip"):
    """Import DI Workspace with a specific marshalling plugin."""
    with open(file_path, 'rb') as file_handle:
        file = {
            'file': (
                os.path.basename(file_path),
                file_handle
            )
        }
        import_uri = get_import_uri().format(plugin_name)
        response = send_request(
            import_uri,
            method="POST",
            files=file
        )
        return response



"""API for working with a single project activity."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_activity_uri():
    """Get endpoint URI pattern for a activity."""
    path = "/workspace/projects/{}/activities/{}"
    return config.get_di_api_endpoint() + path


def get_start_activity_uri():
    """Get start endpoint URI pattern for a activity."""
    return get_activity_uri() + "/start"


def get_start_blocking_activity_uri():
    """Get start-blocking endpoint URI pattern for a activity."""
    return get_activity_uri() + "/startBlocking"


def get_cancel_activity_uri():
    """Get cancel endpoint URI pattern for a activity."""
    return get_activity_uri() + "/cancel"


def get_config_activity_uri():
    """Get config endpoint URI pattern for a activity."""
    return get_activity_uri() + "/config"


def get_status_activity_uri():
    """Get status endpoint URI pattern for a activity."""
    return get_activity_uri() + "/status"


def get_value_activity_uri():
    """Get value endpoint URI pattern for a activity."""
    return get_activity_uri() + "/value"


def start_project_activity(project_name,
                           activity_name,
                           blocking=False,
                           data=None):
    """POST start activity."""
    if blocking:
        return start_blocking_project_activity(
            project_name,
            activity_name,
            data
        )
    return send_request(
        get_start_activity_uri().format(
            project_name,
            activity_name
        ),
        method="POST",
        data=data
    ).decode("utf-8")


def get_project_activity_status(project_name, activity_name):
    """GET project activity status."""
    return json.loads(send_request(
        get_status_activity_uri().format(
            project_name,
            activity_name
        ),
        method="GET"
    ).decode("utf-8"))


def cancel_project_activity(project_name, activity_name):
    """Cancel project activity."""
    return send_request(
        get_cancel_activity_uri().format(
            project_name,
            activity_name
        ),
        method="POST"
    ).decode("utf-8")


def start_blocking_project_activity(project_name, activity_name, data):
    """POST start blocking activity."""
    return send_request(
        get_start_blocking_activity_uri().format(
            project_name,
            activity_name
        ),
        method="POST",
        data=data
    ).decode("utf-8")



"""Export workspace."""
from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_export_uri():
    """Get endpoint URI pattern for workspace export."""
    return config.get_di_api_endpoint() + "/workspace/export/{}"


def export_xml_zip():
    """Export DI Workspace to XML ZIP."""
    return export("xmlZip")


def export_rdf_turtle():
    """Export DI Workspace as RDF Turtle.

    This does not export the resources.
    """
    return export("rdfTurtle")


def export(plugin_name="xmlZip"):
    """Export DI workspace with a specific marshalling plugin.

    Default plugin is xmlZip.
    """
    export_uri = get_export_uri().format(plugin_name)
    export_data = send_request(export_uri)
    return export_data



"""cmempy configuration."""
import os

import certifi


def get_cmem_user_agent():
    """Get the CMEM_USER_AGENT config value.

    The value of user agent is send with each request.
    """
    return os.environ.get(
        "CMEM_USER_AGENT",
        "unknown"
    )


def get_cmem_base_protocol():
    """Get the CMEM_BASE_PROTOCOL config value."""
    return os.environ.get(
        "CMEM_BASE_PROTOCOL",
        "http"
    )


def get_cmem_base_domain():
    """Get the CMEM_BASE_DOMAIN config value."""
    return os.environ.get(
        "CMEM_BASE_DOMAIN",
        "docker.localhost"
    )


def get_cmem_base_uri():
    """Get the CMEM_BASE_URI config value (no ending slash)."""
    return str(
        os.environ.get(
            "CMEM_BASE_URI",
            "%s://%s" % (get_cmem_base_protocol(), get_cmem_base_domain())
        )
        .strip("/")
    )


def get_ssl_verify():
    """Get the SSL_VERIFY config value."""
    ssl_verify = os.environ.get(
        "SSL_VERIFY",
        "True"
    )
    if ssl_verify.lower() == "false":
        ssl_verify = False
    else:
        ssl_verify = True
    return ssl_verify


def get_requests_ca_bundle():
    """Get the REQUESTS_CA_BUNDLE config value."""
    requests_ca_bundle = os.environ.get(
        "REQUESTS_CA_BUNDLE",
        certifi.where()
    )
    requests_ca_bundle = os.path.abspath(requests_ca_bundle)
    if not os.path.isfile(requests_ca_bundle):
        raise ValueError("The value of REQUESTS_CA_BUNDLE does not point to "
                         "a CA bundle file ({}).".format(requests_ca_bundle))
    return requests_ca_bundle


def get_di_api_endpoint():
    """Get the DI_API_ENDPOINT config value."""
    return os.environ.get(
        "DI_API_ENDPOINT",
        "%s/dataintegration" % (get_cmem_base_uri(),)
    )


def get_dp_api_endpoint():
    """Get the DP_API_ENDPOINT config value."""
    return os.environ.get(
        "DP_API_ENDPOINT",
        "%s/dataplatform" % (get_cmem_base_uri(),)
    )


def get_oauth_token_uri():
    """Get the OAUTH_TOKEN_URI config value."""
    return os.environ.get(
        "OAUTH_TOKEN_URI",
        "%s/auth/realms/cmem/protocol/openid-connect/token"
        % (get_cmem_base_uri(),)
    )


def get_oauth_grant_type():
    """Get the OAUTH_GRANT_TYPE config value."""
    valid_values = (
        "password",
        "client_credentials",
        "prefetched_token"
    )
    value = os.environ.get(
        "OAUTH_GRANT_TYPE",
        "client_credentials"
    )
    if value not in valid_values:
        raise ValueError(
            "OAUTH_GRANT_TYPE is set to {} which is not set valid value {}"
            .format(value, valid_values)
        )
    if value == "prefetched_token" and get_oauth_access_token() is None:
        raise ValueError(
            "OAUTH_GRANT_TYPE is set to 'prefetched_token', "
            "but no OAUTH_ACCESS_TOKEN is given."
        )
    return value


def get_oauth_access_token():
    """Get the OAUTH_ACCESS_TOKEN config value."""
    return os.environ.get(
        "OAUTH_ACCESS_TOKEN",
        None
    )


def get_oauth_user():
    """Get the OAUTH_USER config value."""
    default_value = None
    if get_oauth_grant_type() == "password":
        default_value = "admin"
    return os.environ.get(
        "OAUTH_USER",
        default_value
    )


def get_oauth_password():
    """Get the OAUTH_PASSWORD config value."""
    default_value = None
    if get_oauth_grant_type() == "password":
        default_value = "admin"
    return os.environ.get(
        "OAUTH_PASSWORD",
        default_value
    )


def get_oauth_client_id():
    """Get the OAUTH_CLIENT_ID config value."""
    return os.environ.get(
        "OAUTH_CLIENT_ID",
        "cmem-service-account"
    )


def get_oauth_client_secret():
    """Get the OAUTH_CLIENT_SECRET config value."""
    return os.environ.get(
        "OAUTH_CLIENT_SECRET",
        "c8c12828-000c-467b-9b6d-2d6b5e16df4a"
    )


def get_oauth_default_credentials():
    """Get the default credentials object."""
    return {
        "grant_type": get_oauth_grant_type(),
        "username": get_oauth_user(),
        "password": get_oauth_password(),
        "client_id": get_oauth_client_id(),
        "client_secret": get_oauth_client_secret()
    }


def is_chatty():
    """Check if API is in chatty mode. True means additional print outputs."""
    env = os.environ.get(
        "CMEMPY_IS_CHATTY",
        "true"
    )
    if env.lower() == "false":
        # silentio
        return False
    # oratio
    return True


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
DATA_DIR = os.path.join(CURRENT_DIR, "..", "data")



"""Export project."""
from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_export_uri():
    """Get endpoint URI pattern for project export."""
    path = "/workspace/projects/{}/export/{}"
    return config.get_di_api_endpoint() + path


def export_project_rdf_turtle(project_name):
    """Export DI Project as RDF Turtle.

    This does not export the resources.
    """
    return export_project(project_name, "rdfTurtle")


def export_project(project_name, plugin_name="xmlZip"):
    """Export DI Project with a specific marshalling plugin.

    Default plugin is xmlZip.
    """
    export_uri = get_export_uri().format(project_name, plugin_name)
    export_data = send_request(export_uri)
    return export_data



"""Import project."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import (
    get_json,
    request,
    send_request,
)


def get_import_uri():
    """Get endpoint URI pattern for project import."""
    path = "/workspace/projects/{}/import/{}"
    return config.get_di_api_endpoint() + path


def import_project_rdf_turtle(project_name, rdf_turtle_string):
    """Import DI Project from RDF Turtle (without resource files)."""
    import_uri = get_import_uri().format(project_name, "rdfTurtle")
    response = send_request(
        import_uri,
        method="POST",
        data=rdf_turtle_string.encode("utf-8")
    )
    return response


def import_project(project_name, file_path, plugin_name="xmlZip"):
    """Import DI Project from an exported file (deprecated)."""
    import_uri = get_import_uri().format(project_name, plugin_name)
    with open(file_path, 'rb') as data:
        response = send_request(
            import_uri,
            method="POST",
            data=data
        )
    return response


def upload_project(file_path):
    """Upload a project file of the project to be imported.

    Project import based on uploaded resources is a multi step procedure.

    This function implements step 1 and 2 of the following process:

    - (1.) the project file upload,
    - (2.) the validation of the uploaded file,
    - (3.) the asynchronous execution of the project import and
    - (4.) the status of the running project import execution.

    Args:
        file_path (str): path to the ZIP or turtle (depends on plugin)

    Returns
        dict of the validation response
    """
    api_base = config.get_di_api_endpoint()

    # 1. the project file upload
    upload_path = "/api/workspace/projectImport"
    upload_uri = api_base + upload_path
    with open(file_path, 'rb') as data:
        project_import_id = json.loads(
            request(upload_uri, method="POST", data=data)
            .content.decode("utf-8")
        )["projectImportId"]

    # 2. the validation of the uploaded file
    validation_path = "/api/workspace/projectImport/" + project_import_id
    validation_uri = api_base + validation_path
    # projectId, label, marshallerId, projectAlreadyExists
    validation_response = json.loads(
        request(validation_uri, method="GET")
        .content.decode("utf-8")
    )
    validation_response["projectImportId"] = project_import_id
    return validation_response


def import_from_upload_start(
        import_id,
        project_id=None,
        generate_id=False,
        overwrite_existing=False
):
    """Import/create a project from an uploaded file.

    Project import based on uploaded resources is a multi step procedure.

    This function implements step 3 of the following process:

    - (1.) the project file upload,
    - (2.) the validation of the uploaded file,
    - (3.) the asynchronous execution of the project import and
    - (4.) the status of the running project import execution.

    Args:
        import_id (str): The projectUploadID received by upload_project()
        project_id (str): If provided, this will adopt the given id for the
            imported project. Cannot be set together with 'generateNewId'.
        generate_id (bool): When enabled this will always generate a new ID
            for this project based on the project label. This is one strategy
            if a project with the original ID already exists.
        overwrite_existing (bool): When enabled this will overwrite an
            existing project with the same ID. Enabling this option will
            NOT override the generateNewId option.

    Returns
        response object of the request
    """
    api_base = config.get_di_api_endpoint()
    endpoint_path = "/api/workspace/projectImport/" + import_id
    endpoint_uri = api_base + endpoint_path
    params = {
        "overwriteExisting": "true" if overwrite_existing else "false",
        "generateNewId": "true" if generate_id else "false",
    }
    if project_id is not None:
        params["newProjectId"] = project_id
    # 3. the asynchronous execution of the project import
    response = request(endpoint_uri, method="POST", params=params)
    return response


def import_from_upload_status(import_id):
    """Get the status of the import project activity.

    Project import based on uploaded resources is a multi step procedure.

    This function implements step 4 of the following process:

    - (1.) the project file upload,
    - (2.) the validation of the uploaded file,
    - (3.) the asynchronous execution of the project import and
    - (4.) the status of the running project import execution.

    Args:
        import_id (str): The projectUploadID received by upload_project()

    Returns
        status object
    """
    api_base = config.get_di_api_endpoint()
    path = "/api/workspace/projectImport/{}/status".format(import_id)
    endpoint = api_base + path
    # 4. the status of the running project import execution.
    return get_json(endpoint, params={"timeout": "1000"})



"""API methods for working with several projects."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_projects_uri():
    """Get endpoint URI for project list."""
    path = "/workspace/projects"
    return config.get_di_api_endpoint() + path


def get_projects():
    """GET list of projects."""
    response = send_request(get_projects_uri(), method="GET")
    return json.loads(response.decode("utf-8"))



"""API for retrieving marshalling plugins."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request

KNOWN_PLUGINS = {
    "xmlZip": "zip",
    "rdfTurtle": "ttl"
}


def get_marshalling_plugins_uri():
    """Get endpoint URI for the marshalling plugins list."""
    path = "/workspace/marshallingPlugins"
    return config.get_di_api_endpoint() + path


def get_marshalling_plugins():
    """GET retrieve marshalling plugins."""
    return json.loads(send_request(
        get_marshalling_plugins_uri(),
        method="GET"
    ).decode("utf-8"))


def get_marshalling_plugins_as_dict():
    """Retrieve marshalling plugins as extended dictionary."""
    plugins = {}
    for plugin in get_marshalling_plugins():
        plugins[plugin["id"]] = plugin
        if plugin["id"] in KNOWN_PLUGINS.keys()\
                and "fileExtension" not in plugin.keys():
            plugin["fileExtension"] = KNOWN_PLUGINS[plugin["id"]]
    return plugins


def get_extension_by_plugin(plugin_id):
    """Return a filename extension string of a marshalling plugin."""
    plugins = get_marshalling_plugins_as_dict()
    if plugin_id in plugins.keys():
        if "fileExtension" in plugins[plugin_id].keys():
            return plugins[plugin_id]["fileExtension"]
        if plugin_id in KNOWN_PLUGINS.keys():
            return KNOWN_PLUGINS[plugin_id]
        raise ValueError("Could not get file extension for plugin "
                         + plugin_id)
    raise ValueError(plugin_id + " is not a valid marshalling plugin. "
                     + "Try one of " + str(KNOWN_PLUGINS.keys()))


def get_plugin_by_extension(extension):
    """Return a marshalling plugin for a given file extension."""
    for plugin_id, plugin in get_marshalling_plugins_as_dict().items():
        # return the first match of the extension
        if plugin["fileExtension"] == extension:
            return plugin_id
    raise ValueError("Could not get marshalling plugin for file extension "
                     + extension)



"""API for managing conditions in DP."""

from cmem.cmempy import config
from cmem.cmempy.api import request


def get_graph_uri():
    """Get graph URI of the authorization conditions graph."""
    path = "/authorization/conditions"
    return config.get_dp_api_endpoint() + path


def get():
    """GET access conditions."""
    accept = "application/n-triples"
    headers = {"Accept": accept}
    return request(
        get_graph_uri(),
        method="GET",
        headers=headers
    )


def put(
        resource,
        access_conditions,
        rdf_serialization="application/n-triples"):
    """PUT access conditions."""
    params = {
        "resource": resource
    }
    data = access_conditions
    headers = {"Content-Type": rdf_serialization}
    return request(
        get_graph_uri(),
        method="PUT",
        params=params,
        data=data,
        headers=headers
    )


def post(
        access_conditions,
        rdf_serialization="application/n-triples"):
    """POST access conditions."""
    data = access_conditions
    headers = {"Content-Type": rdf_serialization}
    return request(
        get_graph_uri(),
        method="POST",
        data=data,
        headers=headers
    )


def delete(resource):
    """DELETE access condition."""
    params = {
        "resource": resource
    }
    return request(
        get_graph_uri(),
        method="DELETE",
        params=params
    )



#!/usr/bin/env python
"""setup.py for cmempy."""

from setuptools import setup, find_packages

setup(
    name='cmem_cmempy',
    version='22.2',
    author='eccenca',
    author_email='cmempy-developer@eccenca.com',
    maintainer='Sebastian Tramp',
    maintainer_email='sebastian.tramp@eccenca.com',
    url='https://eccenca.com/go/cmempy',
    description='API wrapper for eccenca Corporate Memory',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    license='Apache 2.0',
    install_requires=[
        "certifi",
        "pyparsing",
        "pysocks",
        "rdflib",
        "requests",
        "requests_toolbelt",
        "six"
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Testing',
        'Topic :: Database',
        'Topic :: Utilities',
    ],
)



"""API for working with the vocabulary catalog."""

import json

import requests

from cmem.cmempy import config
from cmem.cmempy.api import send_request
from cmem.cmempy.dp.proxy.graph import (
    delete as delete_graph,
    put as put_graph
)
from cmem.cmempy.queries import SparqlQuery
from cmem.cmempy.workspace import (
    reload_prefixes,
    update_global_vocabulary_cache
)

DEFAULT_CATALOG_GRAPH = "https://ns.eccenca.com/example/data/vocabs/"

REMOVE_CATALOG_ENTRY_IF_NOT_INSTALLABLE = """
PREFIX dcat: <http://www.w3.org/ns/dcat#>
WITH <{{graph}}>
DELETE { ?s ?p ?o }
WHERE {
    ?s ?p ?o .
    FILTER NOT EXISTS { ?s dcat:distribution ?downloadUrl . }
    FILTER (STR(?s) = "{{vocab}}")
}
"""


def get_vocabs_endpoint():
    """Get endpoint URL for the dataplatform vocabs API."""
    return config.get_dp_api_endpoint() + "/api/vocabs"


def get_global_vocabs_cache_endpoint():
    """Get endpoint URL for the dataintegration vocabs cache."""
    path = "/workspace/globalWorkspaceActivities/GlobalVocabularyCache/value"
    return config.get_di_api_endpoint() + path


def get_vocabularies(
        catalog_uri=DEFAULT_CATALOG_GRAPH,
        limit=1000,
        offset=0,
        filter_="all"):
    """Get all vocabularies from a vocabulary catalog."""
    allowed_filter = ("all", "installed", "installable")
    if filter_ not in allowed_filter:
        raise ValueError(
            "Parameter filter_ not in ({})".format(', '.join(allowed_filter))
        )
    params = {
        "vocabGraph": catalog_uri,
        "limit": limit,
        "offset": offset
    }
    response = send_request(
        get_vocabs_endpoint(),
        method="GET",
        params=params
    )
    vocabs = json.loads(response.decode("utf-8"))["vocabularies"]
    if filter_ == "installed":
        vocabs = [
            vocab for vocab in vocabs
            if vocab["installed"]
        ]
    if filter_ == "installable":
        vocabs = [
            vocab for vocab in vocabs
            if not vocab["installed"] and vocab["downloadUrl"]
        ]
    return vocabs


def get_vocabulary(iri, catalog_uri=DEFAULT_CATALOG_GRAPH):
    """Get a vocabulary from a vocabulary catalog."""
    vocabs = get_vocabularies(catalog_uri=catalog_uri)
    for _ in vocabs:
        if _["iri"] == iri:
            return _
    raise FileNotFoundError("not available in the catalog")


def install_vocabulary(iri, catalog_uri=DEFAULT_CATALOG_GRAPH):
    """Install a vocabulary."""
    vocab = get_vocabulary(iri, catalog_uri=catalog_uri)
    if "downloadUrl" not in vocab:
        raise KeyError("no downloadable URL found")
    response = requests.api.request(
        "GET",
        vocab["downloadUrl"],
        verify=True,
    )
    if response.status_code is requests.status_codes.codes.ok:
        response.raise_for_status()
    put_graph(iri, response.content, "text/turtle")
    reload_prefixes()
    update_global_vocabulary_cache(iri)


def uninstall_vocabulary(iri, catalog_uri=DEFAULT_CATALOG_GRAPH):
    """Delete a vocabulary."""
    # make sure its a vocabulary graph
    get_vocabulary(iri, catalog_uri=catalog_uri)
    delete_graph(iri)
    reload_prefixes()
    update_global_vocabulary_cache(iri)
    remove_metadata = SparqlQuery(
        REMOVE_CATALOG_ENTRY_IF_NOT_INSTALLABLE,
        query_type="UPDATE"
    )
    placeholder = {
        "graph": DEFAULT_CATALOG_GRAPH,
        "vocab": iri
    }
    remove_metadata.get_results(placeholder=placeholder)


def get_global_vocabs_cache():
    """Get the global vocabulary cache value."""
    headers = {
        "Accept": "application/json"
    }
    response = send_request(
        get_global_vocabs_cache_endpoint(),
        method="GET",
        headers=headers
    )
    cache_ = json.loads(response.decode("utf-8"))
    return cache_



"""API for working with task activities."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import (
    get_json,
    send_request
)
from cmem.cmempy.workspace.activities import (
    ACTIVITY_TYPE_EXECUTE_LOCALWORKFLOW,
    VALID_ACTIVITY_STATUS,
    taskactivity
)


def get_activities_uri():
    """Get endpoint URI pattern for a activities."""
    path = "/workspace/projects/{}/tasks/{}/activities"
    return config.get_di_api_endpoint() + path


def get_task_activities(project_name, task_name):
    """GET retrieve a list of task activities."""
    response = send_request(
        get_activities_uri().format(project_name, task_name),
        method="GET"
    )
    return json.loads(response.decode("utf-8"))


# Kept here for backward compatibility
# DO NOT USE
def start_task_activity(project_name,
                        task_name,
                        activity_name=ACTIVITY_TYPE_EXECUTE_LOCALWORKFLOW,
                        blocking=False,
                        data=None):
    """POST start task activity."""
    return taskactivity.start_task_activity(
        project_name,
        task_name,
        activity_name=activity_name,
        blocking=blocking,
        data=data
    )


# Kept here for backward compatibility
# DO NOT USE
def get_activity_status(project_name, task_name, activity_name):
    """GET retrieve activity status."""
    return taskactivity.get_activity_status(
        project_name, task_name, activity_name
    )


def get_activities_status(project_name=None, status=None, activity=None):
    """Retrieve status information of activities.

    The corresponding API is new in CMEM 20.10

    If status is defined only task activities with a specific status are
    returned.
    - Valid values are "Idle", "Not executed", "Finished", "Cancelled",
      "Failed", "Successful", "Canceling", "Running" and "Waiting".
    - States "Idle" and "Not executed" are synonyms and "Idle" is kept only for
      backwards compatibility.
    - State "Finished" is a union of following sub-states "Cancelled",
      "Failed" and "Successful".
    - "Waiting" is the state of an activity being scheduled, but still waiting
      in queue for being executed.

    activity name examples: ExecuteLocalWorkflow, ExecuteDefaultWorkflow,
        ExecuteTransform, TypesCache, ...

    Args:
        project_name (str): The DI project ID.
        status (str): One of the documented status names (e.g. "Finished")
        activity (str): get only activities with a specific name

    Raises:
        ValueError: in case the dataset ID is not splittable
    """
    path = "/api/workspace/taskActivitiesStatus"
    task_activities_status_api = config.get_di_api_endpoint() + path
    params = {}
    if project_name:
        params["projectId"] = project_name
    if status:
        if status not in VALID_ACTIVITY_STATUS:
            raise ValueError(
                "Status {} is not a valid activity status, one of {}"
                .format(status, VALID_ACTIVITY_STATUS)
            )
        params["statusFilter"] = status
    status_info = get_json(
        task_activities_status_api,
        params=params
    )
    if activity:
        status_info = [
            info for info in status_info
            if info["activity"] == activity
        ]
    return status_info



"""API for working with task activities."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request
from cmem.cmempy.workspace.activities import (
    ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW
)


def get_activity_uri():
    """Get endpoint URI pattern for an activity."""
    path = "/workspace/projects/{}/tasks/{}/activities/{}"
    return config.get_di_api_endpoint() + path


def get_start_activity_uri():
    """Get start endpoint URI pattern for an activity."""
    return get_activity_uri() + "/start"


def get_start_blocking_activity_uri():
    """Get start-blocking endpoint URI pattern for an activity."""
    return get_activity_uri() + "/startBlocking"


def get_config_activity_uri():
    """Get config endpoint URI pattern for an activity."""
    return get_activity_uri() + "/config"


def get_status_activity_uri():
    """Get status endpoint URI pattern for an activity."""
    return get_activity_uri() + "/status"


def get_value_activity_uri():
    """Get value endpoint URI pattern for an activity."""
    return get_activity_uri() + "/value"


def start_task_activity(project_name,
                        task_name,
                        activity_name=ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW,
                        blocking=False,
                        data=None):
    """POST start task activity."""
    if blocking:
        return start_blocking_task_activity(
            project_name,
            task_name,
            activity_name=activity_name,
            data=data
        )
    return send_request(
        get_start_activity_uri().format(
            project_name,
            task_name,
            activity_name
        ),
        method="POST",
        data=data
    ).decode("utf-8")


def start_blocking_task_activity(
        project_name,
        task_name,
        activity_name=ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW,
        data=None
):
    """POST start blocking task activity."""
    return send_request(
        get_start_blocking_activity_uri().format(
            project_name,
            task_name,
            activity_name
        ),
        method="POST",
        data=data
    ).decode("utf-8")


def get_activity_status(
        project_name,
        task_name,
        activity_name=ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW):
    """GET retrieve activity status."""
    return json.loads(send_request(
        get_status_activity_uri().format(
            project_name,
            task_name,
            activity_name
        ),
        method="GET"
    ).decode("utf-8"))



"""Send HTTP requests to CMEM API."""
import json

import requests

from . import config


def get_access_token(_oauth_token_uri=None, _oauth_credentials=None):
    """Return access token string.

    This is a wrapper for get_token to get only the string.
    It does NOT fetch a new token in case a 'prefetched_token" grant type.
    """
    if config.get_oauth_grant_type() == "prefetched_token":
        return config.get_oauth_access_token()
    return get_token(_oauth_token_uri, _oauth_credentials)["access_token"]


def get_token(_oauth_token_uri=None,
              _oauth_credentials=None):
    """
    Return auth token json structure.

    Auth token format is:
    {
        'token_type': 'bearer',
        'refresh_token': '923a28d6-470d-4afd-9b91-0c952a29a423',
        'expires_in': 43199,
        'access_token': '55c21f4e-9529-4399-aaf1-128acb9da4a1',
        'scope': 'read'
    }
    """
    if not _oauth_credentials:
        _oauth_credentials = config.get_oauth_default_credentials()
    if not _oauth_token_uri:
        _oauth_token_uri = config.get_oauth_token_uri()
    post_req = _request(
        'post',
        _oauth_token_uri,
        data=_oauth_credentials
    )
    post_req.raise_for_status()
    response = post_req.content.decode("utf-8")
    try:
        parsed_json = json.loads(response)
    except json.decoder.JSONDecodeError:
        # pylint: disable=raise-missing-from
        raise ValueError(
            "Token response is not proper JSON. Response was:\n{}"
            .format(response)
        )
    return parsed_json


# pylint: disable-msg=too-many-arguments
def send_request(uri, method="GET", data=None,
                 _json=None, headers=None, params=None,
                 files=None):
    """Return HTML content for request directly."""
    return request(uri, method=method, data=data,
                   _json=_json, headers=headers, params=params,
                   files=files).content


# pylint: disable-msg=too-many-arguments
def get_json(uri, method="GET", data=None, _json=None,
             headers=None, params=None, files=None):
    """Return parsed JSON content for request directly."""
    response = request(uri, method=method, data=data, _json=_json,
                       headers=headers, params=params, files=files).content
    return json.loads(response.decode("utf-8"))


def _request(method=None, url=None, **kwargs):
    return requests.api.request(
        method,
        url,
        verify=config.get_ssl_verify(),
        **kwargs
    )


# pylint: disable-msg=too-many-arguments
def request(uri, method="GET", data=None,
            _json=None, headers=None, params=None,
            files=None, stream=False):
    """Return raw response object."""
    if headers is None:
        headers = {}
    headers['Authorization'] = "Bearer {}".format(str(get_access_token()))
    headers['User-Agent'] = config.get_cmem_user_agent()
    if method == "GET":
        response = _request(
            'get',
            uri,
            headers=headers,
            stream=stream,
            params=params
        )
    elif method == "POST":
        response = _request(
            'post',
            uri,
            headers=headers,
            data=data,
            params=params,
            files=files,
            stream=stream
        )
    elif method == "PUT":
        if config.is_chatty():
            print(headers)
        response = _request(
            'put',
            uri,
            headers=headers,
            data=data,
            params=params,
            files=files,
            stream=stream
        )
    elif method == "DELETE":
        response = _request(
            'delete',
            uri,
            headers=headers,
            params=params,
            stream=stream
        )
    elif method == "PATCH":
        response = _request(
            'patch',
            uri,
            headers=headers,
            data=data,
            params=params,
            files=files,
            stream=stream
        )
    else:
        raise Exception("Unknown HTTP request method")
    if response.status_code not in [requests.status_codes.codes.ok,
                                    requests.status_codes.codes.no_content,
                                    requests.status_codes.codes.created]:
        if config.is_chatty():
            print("Request failed for URL: {} {}".format(
                response.url, response.request.method))
            print(response.content.decode("utf-8"))
            print(response.status_code)
    response.raise_for_status()
    return response



"""API methods for working with single datasets."""
import json

from cmem.cmempy import config
from cmem.cmempy.api import send_request


def get_dataset_uri():
    """Get endpoint URI pattern for a dataset."""
    path = "/workspace/projects/{}/datasets/{}"
    return config.get_di_api_endpoint() + path


def get_dataset(project_name, dataset_name):
    """GET retrieve single dataset."""
    headers = {
        "Accept": "application/json"
    }
    response = send_request(
        get_dataset_uri().format(project_name, dataset_name),
        method="GET",
        headers=headers
    )
    return json.loads(response.decode("utf-8"))


def make_new_dataset(project_name, dataset_name,
                     dataset_type, parameters, autoconfigure):
    """PUT create new dataset (deprecated)."""
    data = {
        "id": "{}".format(dataset_name),
        "type": "{}".format(dataset_type),
        "parameters": parameters
    }
    data = json.dumps(data).encode("utf-8")
    headers = {"Content-Type": "application/json"}
    params = None
    if autoconfigure:
        params = {"autoConfigure": "true"}
    send_request(
        get_dataset_uri().format(project_name, dataset_name),
        method="PUT",
        data=data,
        headers=headers,
        params=params
    )


def delete_dataset(project_name, dataset_name):
    """DELETE remove existing dataset."""
    send_request(
        get_dataset_uri().format(project_name, dataset_name),
        method="DELETE"
    )


def create_dataset(
        project_id, dataset_type,
        dataset_id=None, parameter=None, metadata=None
):
    """Create a dataset.

    In difference to make_new_dataset, this uses the task API and does
    not enforce an ID in advance. Also it allows for metadata parameters.
    """
    if parameter is None:
        parameter = {}
    if metadata is None:
        metadata = {}
    # add needed base task data
    data = {
        "taskType": "Dataset",
        "type": dataset_type,
        "parameters": parameter,
        "metadata": metadata
    }
    # add the optional task / dataset ID
    if dataset_id:
        data["id"] = dataset_id
    params = None
    path = "/workspace/projects/{}/tasks".format(project_id)
    return send_request(
        config.get_di_api_endpoint() + path,
        method="POST",
        data=json.dumps(data).encode("utf-8"),
        headers={"Content-Type": "application/json"},
        params=params
    )
